import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';

class CandlesticksChart extends StatefulWidget {
  final Size chartSize;
  final int minimumScale, maximumScale;
  final List<CandleModel> data;
  final Color backgroundColor;

  const CandlesticksChart(
      {Key key,
      this.chartSize = const Size(double.infinity, 300),
      this.minimumScale,
      this.maximumScale,
      @required this.data,
      this.backgroundColor = const Color.fromARGB(255, 18, 32, 47)})
      : super(key: key);

  @override
  _CandlesticksChartState createState() => _CandlesticksChartState(
        data,
        minimumScale: minimumScale ?? 1,
        maximumScale: maximumScale ?? 2,
      );
}

class _CandlesticksChartState extends State<CandlesticksChart> {
  Offset mousePointer;
  GlobalKey _key = GlobalKey();
  Size size;
  Offset position;
  int lastItemIndex; //data items length
  int scale;

  final int minimumScale, maximumScale;
  final List<CandleModel> data;

  _CandlesticksChartState(this.data, {this.minimumScale, this.maximumScale}) {
    scale = maximumScale;
    lastItemIndex = data.length;
  }

  @override
  void initState() {
    WidgetsBinding.instance.addPostFrameCallback((_) {
      size = _getSize(_key);
      position = _getPosition(_key);
    });

    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Container(
      key: _key,
      height: widget.chartSize.height,
      width: widget.chartSize.width,
      color: widget.backgroundColor,
      child: MouseRegion(
        onHover: (e) {
          setState(() {
            mousePointer = e.position - position;
          });
        },
        onExit: (e) {
          setState(() {
            mousePointer = null;
          });
        },
        child: Listener(
          onPointerSignal: (pointerSignal) {
            if (pointerSignal is PointerScrollEvent) {
              _scaleDownOrUp(
                  pointerSignal.scrollDelta, pointerSignal.localPosition);
            }
          },
          child: GestureDetector(
//            onVerticalDragUpdate: (DragUpdateDetails u) {
//              _scaleDownOrUp(u.delta, u.localPosition);
//            },
            onHorizontalDragUpdate: (DragUpdateDetails u) {
              setState(() {
                mousePointer = u.localPosition;
                if (u.delta.dx > 3) {
                  if (lastItemIndex > 1) lastItemIndex--;
                } else if (u.delta.dx < -3) {
                  if (lastItemIndex < data.length) lastItemIndex++;
                }
              });
            },
            onHorizontalDragEnd: (DragEndDetails e) {
              setState(() {
                mousePointer = null;
              });
            },
            child: CustomPaint(
              painter: Chart(data,
                  showingItem: lastItemIndex,
                  enableGridLines: true,
                  increaseColor: Colors.white,
                  decreaseColor: Colors.blueAccent,
                  mousePointer: mousePointer,
                  scale: scale),
            ),
          ),
        ),
      ),
    );
  }

  void _scaleDownOrUp(Offset offset, Offset position) {
    setState(() {
      mousePointer = position;
      if (offset.dy > 3) {
        if (scale >= minimumScale && scale < maximumScale) scale++;
      } else if (offset.dy < -3) {
        if (scale > minimumScale && scale <= maximumScale) scale--;
      }
    });
  }

  Offset _getPosition(GlobalKey key) {
    final RenderBox box = key.currentContext.findRenderObject();
    final Offset position = box.localToGlobal(Offset.zero);
    return position;
  }

  Size _getSize(GlobalKey key) {
    final RenderBox box = key.currentContext.findRenderObject();
    return box.size;
  }
}

class CandleModel {
  final double high;
  final double low;
  final double open;
  final double close;
  final double volumeTo;

  CandleModel({this.high, this.low, this.open, this.close, this.volumeTo = 10});

  @override
  String toString() {
    return "ID=>${this.volumeTo}, Low: ${this.low}, High: ${this.high}";
  }
}

class Chart extends CustomPainter {
  Chart(this.candles,
      {this.mousePointer,
      this.lineWidth = 1.0,
      this.enableGridLines = false,
      this.gridLineColor = Colors.grey,
      this.gridLineAmount = 6,
      this.gridLineWidth = 0.5,
      this.showingItem = 0,
      this.gridLineLabelColor = Colors.grey,
      this.scale = 10,
      this.labelPrefix = "\$",
      this.increaseColor = Colors.green,
      this.decreaseColor = Colors.red});

  final List<CandleModel> candles;
  final double lineWidth;
  final bool enableGridLines;
  final Color gridLineColor;
  final int gridLineAmount;
  final double gridLineWidth;
  final Color gridLineLabelColor;
  final String labelPrefix;
  final int showingItem, scale;
  final Color increaseColor;
  final Color decreaseColor;
  final Offset mousePointer;

  List<CandleModel> data;

  double _min;
  double _max;

  List<TextPainter> gridLineTextPainters = [];

  numCommaParse(number) {
    return number.round().toString().replaceAllMapped(
        new RegExp(r'(\d{1,3})(?=(\d{3})+(?!\d))'), (Match m) => "${m[1]},");
  }

  update() {
    _min = double.infinity;
    _max = -double.infinity;

    for (var i in data) {
      if (i.high > _max) {
        _max = i.high.toDouble();
      }
      if (i.low < _min) {
        _min = i.low.toDouble();
      }
    }

    if (enableGridLines) {
      double gridLineValue;
      for (int i = 0; i < gridLineAmount; i++) {
        // Label grid lines
        gridLineValue = _max - (((_max - _min) / (gridLineAmount - 1)) * i);

        String gridLineText;
        if (gridLineValue < 1) {
          gridLineText = gridLineValue.toStringAsPrecision(4);
        } else if (gridLineValue < 999) {
          gridLineText = gridLineValue.toStringAsFixed(2);
        } else {
          gridLineText = gridLineValue.round().toString().replaceAllMapped(
              new RegExp(r'(\d{1,3})(?=(\d{3})+(?!\d))'),
              (Match m) => "${m[1]},");
        }

        gridLineTextPainters.add(new TextPainter(
            text: new TextSpan(
                text: labelPrefix + gridLineText,
                style: new TextStyle(
                    color: gridLineLabelColor,
                    fontSize: 10.0,
                    fontWeight: FontWeight.bold)),
            textDirection: TextDirection.ltr));
        gridLineTextPainters[i].layout();
      }
    }
  }

  @override
  void paint(Canvas canvas, Size size) {
    int endIndex = showingItem < candles.length && showingItem > 0
        ? showingItem
        : candles.length;
    int startIndex = endIndex - scale;
    data = candles.sublist(
        startIndex > 0 ? startIndex : 0, startIndex > 0 ? endIndex : scale);

    if (_min == null || _max == null) {
      // || _maxVolume == null) {
      update();
    }

//     final double volumeHeight = size.height * volumeProp;
//     final double volumeNormalizer = volumeHeight / _maxVolume;

    double width = size.width;
    final double height = size.height; // * (1 - volumeProp);

    if (enableGridLines) {
      width = size.width -
          (gridLineTextPainters[0].text as TextSpan).text.length * 6;
      Paint gridPaint = new Paint()
        ..color = gridLineColor
        ..strokeWidth = gridLineWidth;

      double gridLineDist = height / (gridLineAmount - 1);
      double gridLineY;

      // Draw grid lines
      for (int i = 0; i < gridLineAmount; i++) {
        gridLineY = (gridLineDist * i).round().toDouble();
        canvas.drawLine(new Offset(0.0, gridLineY),
            new Offset(width, gridLineY), gridPaint);

        // Label grid lines
        gridLineTextPainters[i]
            .paint(canvas, new Offset(width + 2.0, gridLineY - 6.0));
      }

      // Label volume line
//       maxVolumePainter.paint(canvas, new Offset(0.0, gridLineY + 2.0));
    }

    final double heightNormalizer = height / (_max - _min);
    final double rectWidth = width / data.length;

    double rectLeft;
    double rectTop;
    double rectRight;
    double rectBottom;

    Paint rectPaint;

    // Loop through all data
    for (int i = 0; i < data.length; i++) {
      rectLeft = (i * rectWidth) + lineWidth / 2;
      rectRight = ((i + 1) * rectWidth) - lineWidth / 2;

      rectTop = height - (data[i].open - _min) * heightNormalizer;
      rectBottom = height - (data[i].close - _min) * heightNormalizer;
      rectPaint = Paint()
        ..color = data[i].open > data[i].close ? decreaseColor : increaseColor
        ..strokeWidth = lineWidth
        ..strokeCap = StrokeCap.round
        ..strokeJoin = StrokeJoin.round
        ..strokeMiterLimit = 10;

      Rect ocRect = new Rect.fromLTRB(rectLeft, rectTop, rectRight, rectBottom);
      canvas.drawRect(ocRect, rectPaint);

      // Draw low/high candlestick wicks
      double low = height - (data[i].low - _min) * heightNormalizer;
      double high = height - (data[i].high - _min) * heightNormalizer;
      canvas.drawLine(
          new Offset(rectLeft + rectWidth / 2 - lineWidth / 2, rectBottom),
          new Offset(rectLeft + rectWidth / 2 - lineWidth / 2, low),
          rectPaint);
      canvas.drawLine(
          new Offset(rectLeft + rectWidth / 2 - lineWidth / 2, rectTop),
          new Offset(rectLeft + rectWidth / 2 - lineWidth / 2, high),
          rectPaint);
    }

    if (mousePointer != null) {
      if (mousePointer.dx <= width &&
          mousePointer.dy <= height &&
          mousePointer.dy > 0 &&
          mousePointer.dx > 0) {
        Paint cursorPaint = new Paint()
          ..color = Colors.amber
          ..strokeWidth = 1;

        _paintHorizontalDashedLine(
            0, mousePointer.dy, width, canvas, cursorPaint);

        final chunk = width / data.length;
        final start = (mousePointer.dx / chunk).floor() * chunk;
        final end = start + chunk;
        final center = (start + end) / 2;
        _paintVerticalDashedLine(center, 0, height, canvas, cursorPaint);

        canvas.drawCircle(Offset(center, mousePointer.dy), 2, cursorPaint);

        final amountWidth = size.width - width + 5;
        final amountHeight = 15.0;

        canvas.drawRRect(
            RRect.fromRectAndRadius(
                Rect.fromLTWH(width - 5, mousePointer.dy - amountHeight / 2,
                    amountWidth, amountHeight),
                Radius.circular(4.0)),
            cursorPaint);

        final amount =
            ((height - mousePointer.dy) * (_max - _min) / height) + _min;
        TextSpan span = TextSpan(
            style: new TextStyle(
                fontSize: 10.0,
                fontWeight: FontWeight.bold,
                color: Colors.black),
            text: amount.toStringAsFixed(4));
        TextPainter tp = TextPainter(
            text: span,
            textAlign: TextAlign.left,
            textDirection: TextDirection.ltr);
        tp.layout();
        tp.paint(canvas,
            Offset(width - 2.5, mousePointer.dy - amountHeight / 2 + 2.5));

        _paintDescription(
            width, height, data[(mousePointer.dx / chunk).floor()], canvas);
      }
    }

    _paintGuide(width, height, canvas);
  }

  _paintDescription(
      double width, double height, CandleModel candle, Canvas canvas) {
    String openCloseText = 'Open : ' +
        candle.open.toStringAsFixed(4) +
        ', Close : ' +
        candle.close.toStringAsFixed(4);
    TextSpan span = TextSpan(
        style: new TextStyle(fontSize: 10.0, color: Colors.white),
        text: openCloseText);
    TextPainter tp = TextPainter(
        text: span,
        textAlign: TextAlign.left,
        textDirection: TextDirection.ltr);
    tp.layout();
    tp.paint(canvas, Offset(10, height + 20));

    String highLowText = 'High : ' +
        candle.high.toStringAsFixed(4) +
        ', Low : ' +
        candle.low.toStringAsFixed(4);
    span = TextSpan(
        style: new TextStyle(fontSize: 10.0, color: Colors.white),
        text: highLowText);
    tp = TextPainter(
        text: span,
        textAlign: TextAlign.left,
        textDirection: TextDirection.ltr);
    tp.layout();
    tp.paint(canvas, Offset(6 * openCloseText.length.toDouble(), height + 20));
  }

  _paintGuide(double width, double height, Canvas canvas) {
    var guidePaint = Paint()
      ..strokeWidth = 2
      ..color = decreaseColor;
    canvas.drawRect(Rect.fromLTWH(10, 10, 10, 10), guidePaint);
    TextSpan span = TextSpan(
        style: new TextStyle(fontSize: 10.0, color: Colors.white),
        text: 'Decrease');
    TextPainter tp = TextPainter(
        text: span,
        textAlign: TextAlign.left,
        textDirection: TextDirection.ltr);
    tp.layout();
    tp.paint(canvas, Offset(30, 10));

    guidePaint.color = increaseColor;
    canvas.drawRect(Rect.fromLTWH(10, 30, 10, 10), guidePaint);
    span = TextSpan(
        style: new TextStyle(fontSize: 10.0, color: Colors.white),
        text: 'Increase');
    tp = TextPainter(
        text: span,
        textAlign: TextAlign.left,
        textDirection: TextDirection.ltr);
    tp.layout();
    tp.paint(canvas, Offset(30, 30));
  }

  _paintVerticalDashedLine(
      double x, double y, double max, Canvas canvas, Paint paint) {
    var dashWidth = 5;
    var dashSpace = 5;
    while (max >= 0) {
      canvas.drawLine(Offset(x, y), Offset(x, y + dashWidth), paint);
      final space = (dashSpace + dashWidth);
      y += space;
      max -= space;
    }
  }

  _paintHorizontalDashedLine(
      double x, double y, double max, Canvas canvas, Paint paint) {
    var dashWidth = 5;
    var dashSpace = 5;
    while (max >= 0) {
      canvas.drawLine(Offset(x, y), Offset(x + dashWidth, y), paint);
      final space = (dashSpace + dashWidth);
      x += space;
      max -= space;
    }
  }

  @override
  bool shouldRepaint(Chart old) {
    return data != old.data ||
        lineWidth != old.lineWidth ||
        enableGridLines != old.enableGridLines ||
        gridLineColor != old.gridLineColor ||
        gridLineAmount != old.gridLineAmount ||
        gridLineWidth != old.gridLineWidth ||
        gridLineLabelColor != old.gridLineLabelColor;
  }
}
